.PHONY: all clean

###############################################################################
# Readme
# 
# This makefile assumes $(CFML) is defined.
#
# Optionnally, $(OCAML_FLAGS) can be set (e.g. to "-rectypes").
# Optionnally, $(ML) can be used to specified the sources files (default: *.ml)
# 
# For each %.ml file in $(ML), we create:
# - the files %.d      (dependency files)
# - the files %.cmj    (type information; .cmi files in disguise)
# - the files %_ml.v   (characteristic formulae)
# - the files %_ml.vo  (compiled characteristic formulae)
# 
# Idea number 1:
# Because the CFML generator produces the files %_ml.v and %.cmj at
# the same time (in just one invocation), it is not necessary to
# tell "make" about both of these files. We adopt the convention
# that only the .cmj file is used as a target. This seems to help.
# 
# Idea number 2:
# This Makefile deals only with the generation of %_ml.v and %.cmj.
# Running Coq on %_ml.v is the job of another Makefile, which is
# run in a second stage. This means we write fewer Makefiles, they
# are really independent, and dependency computations are simpler.

###############################################################################
# Parameters

ifndef OCAML_FLAGS
	OCAML_FLAGS := 
	# e.g. -rectypes
endif

ifndef ML
	ML := $(wildcard *.ml)
endif

###############################################################################
# Definitions

CMJ   := $(patsubst %.ml,%.cmj,$(ML))
MLV   := $(patsubst %.ml,%_ml.v,$(ML))
D     := $(patsubst %.ml,%.d,$(ML))

ifeq ($(findstring $(MAKECMDGOALS),clean),)
   include $(CFML)/lib/make/Makefile.tools
endif

OCAMLDEP  := $(OCAMLBIN)ocamldep
OCAMLPOST := $(CFML)/lib/make/ocamldep.post

###############################################################################
# Targets

all: $(CMJ)

###############################################################################
# Dependencies

# As described by the recipe below, including the dependency files $(D)
# adds dependencies of the form:
#   A.cmj: B.cmj
# whenever module A depends on module B.

ifeq ($(findstring $(MAKECMDGOALS),clean),)
-include $(D)
endif

# We use ocamldep -modules to find out which modules A depends upon.
#   (An advantage of ocamldep -modules is that its output does not
#    depend on the presence or absence of certain .mli files.)
#   (A disadvantage of ocamldep -modules is that it does not filter
#    out the standard library modules.)
# We use the script $(OCAMLPOST) to post-process the output of ocamldep
# and filter out the standard library modules.

%.d: %.ml
	@for B in `$(OCAMLDEP) -modules $< | CFML=$(CFML) $(OCAMLPOST)` ; do \
	  echo $*.cmj: $${B}.cmj ; \
	done > $@

###############################################################################
# Rules

# We use CFML's cmj generator to produce %_ml.v and %.cmj files out of %.ml.
# Only the %.cmj target is known to "make".

%.cmj: %.ml $(CFML_MLV)
	$(CFML_MLV) $(OCAML_FLAGS) -I $(CFML)/lib/ocaml -I . $<

###############################################################################
# Clean

clean:
	rm -f *.cmj *_ml.v *.d
	rm -rf output
